@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2022 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2025 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -35,16 +35,20 @@
  *   forward this exception.
  */
 
-#include "common.h"
 #include "prototypes.h"
 
 /* Uncomment to see allocation sources in core dumps */
 /* #define DEBUG_PADDING 64 */
 
+#ifdef HAVE_MIMALLOC_H
+#include <mimalloc.h>
+#define system_malloc(n) mi_malloc(n)
+#define system_realloc(p,n) mi_realloc((p),(n))
+#define system_free(p) mi_free(p)
+#elif defined(USE_WIN32)
 /* reportedly, malloc does not always return 16-byte aligned addresses
  * for 64-bit targets as specified by
  * https://msdn.microsoft.com/en-us/library/6ewkz86d.aspx */
-#ifdef USE_WIN32
 #define system_malloc(n) _aligned_malloc((n),16)
 #define system_realloc(p,n) _aligned_realloc((p),(n),16)
 #define system_free(p) _aligned_free(p)
@@ -61,6 +65,7 @@
 
 /* most platforms require allocations to be aligned */
 #ifdef _MSC_VER
+#pragma warning(disable: 4324)
 __declspec(align(16))
 #endif
 struct alloc_list_struct {
@@ -84,14 +89,19 @@ struct alloc_list_struct {
 
 #define LEAK_TABLE_SIZE 997
 typedef struct {
+    int num, max;   /* current and highest number of allocations */
+    int64_t total; /* approximate total number of heap operations */
     const char *alloc_file;
     int alloc_line;
-    int num, max;
 } LEAK_ENTRY;
 NOEXPORT LEAK_ENTRY leak_hash_table[LEAK_TABLE_SIZE],
     *leak_results[LEAK_TABLE_SIZE];
 NOEXPORT int leak_result_num=0;
 
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+DEFINE_STACK_OF(LEAK_ENTRY)
+#endif /* OpenSSL version >= 1.1.1 */
+
 #ifdef USE_WIN32
 NOEXPORT LPTSTR str_vtprintf(LPCTSTR, va_list);
 #endif /* USE_WIN32 */
@@ -102,8 +112,15 @@ NOEXPORT ALLOC_LIST *get_alloc_list_ptr(void *, const char *, int);
 NOEXPORT void str_leak_debug(const ALLOC_LIST *, int);
 
 NOEXPORT LEAK_ENTRY *leak_search(const ALLOC_LIST *);
-NOEXPORT void leak_report();
-NOEXPORT long leak_threshold();
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+NOEXPORT int leak_cmp(const LEAK_ENTRY *const *, const LEAK_ENTRY *const *);
+#endif /* OpenSSL version >= 1.1.1 */
+NOEXPORT void leak_report(void);
+NOEXPORT long leak_threshold(void);
+
+#if OPENSSL_VERSION_NUMBER<0x10100000L
+NOEXPORT void free_function(void *);
+#endif
 
 TLS_DATA *ui_tls;
 NOEXPORT uint8_t canary[10]; /* 80-bit canary value */
@@ -111,14 +128,6 @@ NOEXPORT volatile uint64_t canary_initialized=CANARY_UNINTIALIZED;
 
 /**************************************** string manipulation functions */
 
-#ifndef va_copy
-#ifdef __va_copy
-#define va_copy(dst, src) __va_copy((dst), (src))
-#else /* __va_copy */
-#define va_copy(dst, src) memcpy(&(dst), &(src), sizeof(va_list))
-#endif /* __va_copy */
-#endif /* va_copy */
-
 char *str_dup_debug(const char *str, const char *file, int line) {
     char *retval;
 
@@ -157,7 +166,7 @@ char *str_printf(const char *format, ...) {
 #endif /* __GNUC__ */
 char *str_vprintf(const char *format, va_list start_ap) {
     int n;
-    size_t size=32;
+    size_t size=96;
     char *p;
     va_list ap;
 
@@ -165,6 +174,7 @@ char *str_vprintf(const char *format, va_list start_ap) {
     for(;;) {
         va_copy(ap, start_ap);
         n=vsnprintf(p, size, format, ap);
+        va_end(ap);
         if(n>-1 && n<(int)size)
             return p;
         if(n>-1)                /* glibc 2.1 */
@@ -213,18 +223,29 @@ NOEXPORT LPTSTR str_vtprintf(LPCTSTR format, va_list start_ap) {
 
 /**************************************** memory allocation wrappers */
 
-void str_init(TLS_DATA *tls_data) {
+void str_init(void) {
+    memset(leak_hash_table, 0, sizeof leak_hash_table);
+#if OPENSSL_VERSION_NUMBER>=0x10100000L
+    CRYPTO_set_mem_functions(str_alloc_detached_debug,
+        str_realloc_detached_debug, str_free_debug);
+#else
+    CRYPTO_set_mem_ex_functions(str_alloc_detached_debug,
+        str_realloc_detached_debug, free_function);
+#endif
+}
+
+void str_thread_init(TLS_DATA *tls_data) {
     tls_data->alloc_head=NULL;
     tls_data->alloc_bytes=tls_data->alloc_blocks=0;
 }
 
-void str_cleanup(TLS_DATA *tls_data) {
+void str_thread_cleanup(TLS_DATA *tls_data) {
     /* free all attached allocations */
     while(tls_data->alloc_head) /* str_free macro requires an lvalue */
         str_free_expression(tls_data->alloc_head+1);
 }
 
-void str_canary_init() {
+void str_canary_init(void) {
     if(canary_initialized!=CANARY_UNINTIALIZED)
         return; /* prevent double initialization on config reload */
     RAND_bytes(canary, (int)sizeof canary);
@@ -233,7 +254,7 @@ void str_canary_init() {
     canary_initialized=CANARY_INITIALIZED; /* after RAND_bytes */
 }
 
-void str_stats() {
+void str_stats(void) {
     TLS_DATA *tls_data;
     ALLOC_LIST *alloc_list;
     int i=0;
@@ -394,6 +415,19 @@ void str_detach_debug(void *ptr, const char *file, int line) {
     }
 }
 
+/* Here be dragons:
+ * this is an ugly hack to work around OpenSSL API that requires const
+ * function parameters -> use it with extreme caution */
+void str_detach_const_debug(const void *ptr, const char *file, int line) {
+    union {
+        const void *c;
+        void *v;
+    } u;
+
+    u.c=ptr;
+    str_detach_debug(u.v, file, line);
+}
+
 void str_free_debug(void *ptr, const char *file, int line) {
     ALLOC_LIST *alloc_list;
 
@@ -419,6 +453,19 @@ void str_free_debug(void *ptr, const char *file, int line) {
     system_free(alloc_list);
 }
 
+/* Here be dragons:
+ * this is an ugly hack to work around OpenSSL API that requires const
+ * function parameters -> use it with extreme caution */
+void str_free_const_debug(const void *ptr, const char *file, int line) {
+    union {
+        const void *c;
+        void *v;
+    } u;
+
+    u.c=ptr;
+    str_free_debug(u.v, file, line);
+}
+
 NOEXPORT ALLOC_LIST *get_alloc_list_ptr(void *ptr, const char *file, int line) {
     ALLOC_LIST *alloc_list;
 
@@ -472,20 +519,24 @@ NOEXPORT void str_leak_debug(const ALLOC_LIST *alloc_list, int change) {
         CRYPTO_THREAD_unlock(stunnel_locks[LOCK_LEAK_HASH]);
     }
 
-    /* for performance we try to avoid calling CRYPTO_atomic_add() here */
+    /* for performance reasons, we ignore the race condition, as an approximate
+     * number of allocations is good enough to identify the most used entries */
+    entry->total++;
+
+    /* for performance reasons, we try to avoid calling CRYPTO_atomic_add() */
 #ifdef USE_OS_THREADS
-#if defined(__GNUC__) && defined(__ATOMIC_ACQ_REL)
+#ifdef _MSC_VER
+    /* casting is safe, because sizeof(long)==sizeof(int) on Windows */
+    allocations=InterlockedExchangeAdd((long *)&entry->num, change)+change;
+#else /* defined(_MSC_VER) */
+#ifdef __ATOMIC_ACQ_REL
     if(__atomic_is_lock_free(sizeof entry->num, &entry->num))
         allocations=__atomic_add_fetch(&entry->num, change, __ATOMIC_ACQ_REL);
-    else
+    else /* atomic add not directly supported by the compiler */
+#endif /* defined(__ATOMIC_ACQ_REL) */
         CRYPTO_atomic_add(&entry->num, change, &allocations,
             stunnel_locks[LOCK_LEAK_HASH]);
-#elif defined(_MSC_VER)
-    allocations=InterlockedExchangeAdd(&entry->num, change)+change;
-#else /* atomic add not directly supported by the compiler */
-    CRYPTO_atomic_add(&entry->num, change, &allocations,
-        stunnel_locks[LOCK_LEAK_HASH]);
-#endif
+#endif /* defined(_MSC_VER) */
 #else /* USE_OS_THREADS */
     allocations=(entry->num+=change);
 #endif /* USE_OS_THREADS */
@@ -510,7 +561,7 @@ NOEXPORT void str_leak_debug(const ALLOC_LIST *alloc_list, int change) {
 /* O(1) hash table lookup */
 NOEXPORT LEAK_ENTRY *leak_search(const ALLOC_LIST *alloc_list) {
     /* a trivial hash based on source file name *address* and line number */
-    unsigned i=((unsigned)(uintptr_t)alloc_list->alloc_file+
+    unsigned i=(1777*(unsigned)(uintptr_t)alloc_list->alloc_file+
         (unsigned)alloc_list->alloc_line)%LEAK_TABLE_SIZE;
 
     while(!(leak_hash_table[i].alloc_line==0 ||
@@ -520,24 +571,56 @@ NOEXPORT LEAK_ENTRY *leak_search(const ALLOC_LIST *alloc_list) {
     return leak_hash_table+i;
 }
 
-void leak_table_utilization() {
+void leak_table_utilization(void) {
     int i, utilization=0;
+    int64_t grand_total=0;
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+    STACK_OF(LEAK_ENTRY) *stats;
+#endif /* OpenSSL version >= 1.1.1 */
 
-    for(i=0; i<LEAK_TABLE_SIZE; ++i) {
-        if(leak_hash_table[i].alloc_line) {
+    /* leak_hash_table[] is only filled at the DEBUG logging level */
+    if(service_options.log_level<LOG_DEBUG)
+        return;
+
+    /* log total hash table utilization */
+    for(i=0; i<LEAK_TABLE_SIZE; ++i)
+        if(leak_hash_table[i].total) {
             ++utilization;
-#if 0
-            s_log(LOG_DEBUG, "Leak hash entry %d: %s:%d", i,
-                leak_hash_table[i].alloc_file, leak_hash_table[i].alloc_line);
-#endif
+            grand_total+=leak_hash_table[i].total;
         }
-    }
-    s_log(LOG_DEBUG, "Leak detection table utilization: %d/%d, %02.2f%%",
+    s_log(LOG_DEBUG, "Leak detection table utilization: %d/%d (%05.2f%%)",
         utilization, LEAK_TABLE_SIZE, 100.0*utilization/LEAK_TABLE_SIZE);
+
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+    /* log up to 5 most frequently used heap allocations */
+    stats=sk_LEAK_ENTRY_new_reserve(leak_cmp, utilization);
+    for(i=0; i<LEAK_TABLE_SIZE; ++i)
+        if(leak_hash_table[i].total)
+            sk_LEAK_ENTRY_push(stats, leak_hash_table + i);
+    sk_LEAK_ENTRY_sort(stats);
+    for(i=0; i<5 && sk_LEAK_ENTRY_num(stats); ++i) {
+        LEAK_ENTRY *entry=sk_LEAK_ENTRY_pop(stats);
+        s_log(LOG_DEBUG, "#%d: %05.2f%% of heap operations: %s:%d",
+            i+1, 100.0*(double)entry->total/(double)grand_total,
+            entry->alloc_file, entry->alloc_line);
+    }
+    sk_LEAK_ENTRY_free(stats);
+#endif /* OpenSSL version >= 1.1.1 */
 }
 
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
+NOEXPORT int leak_cmp(const LEAK_ENTRY *const *a, const LEAK_ENTRY *const *b) {
+    int64_t d = (*a)->total - (*b)->total;
+    if(d>0)
+        return 1;
+    if(d<0)
+        return -1;
+    return 0;
+}
+#endif /* OpenSSL version >= 1.1.1 */
+
 /* report identified leaks */
-NOEXPORT void leak_report() {
+NOEXPORT void leak_report(void) {
     int i;
     long limit;
 
@@ -553,7 +636,7 @@ NOEXPORT void leak_report() {
     CRYPTO_THREAD_unlock(stunnel_locks[LOCK_LEAK_RESULTS]);
 }
 
-NOEXPORT long leak_threshold() {
+NOEXPORT long leak_threshold(void) {
     long limit;
 
     limit=10000*((int)number_of_sections+1);
@@ -574,27 +657,38 @@ int safe_memcmp(const void *s1, const void *s2, size_t n) {
     typedef unsigned long TL;
 #endif
     typedef unsigned char TS;
-    TL r=0, *pl1, *pl2;
-    TS *ps1, *ps2;
+    TL r=0;
+    const TL *pl1, *pl2;
+    const TS *ps1, *ps2;
     int n1=(int)((uintptr_t)s1&(sizeof(TL)-1)); /* unaligned bytes in s1 */
     int n2=(int)((uintptr_t)s2&(sizeof(TL)-1)); /* unaligned bytes in s2 */
 
     if(n1 || n2) { /* either pointer unaligned */
-        ps1=(TS *)s1;
-        ps2=(TS *)s2;
+        ps1=(const TS *)s1;
+        ps2=(const TS *)s2;
     } else { /* both pointers aligned -> compare full words */
-        pl1=(TL *)s1;
-        pl2=(TL *)s2;
+        pl1=(const TL *)s1;
+        pl2=(const TL *)s2;
         while(n>=sizeof(TL)) {
             n-=sizeof(TL);
             r|=(*pl1++)^(*pl2++);
         }
-        ps1=(TS *)pl1;
-        ps2=(TS *)pl2;
+        ps1=(const TS *)pl1;
+        ps2=(const TS *)pl2;
     }
     while(n--)
         r|=(*ps1++)^(*ps2++);
     return r!=0;
 }
 
+/**************************************** OpenSSL allocator hook */
+
+#if OPENSSL_VERSION_NUMBER<0x10100000L
+NOEXPORT void free_function(void *ptr) {
+    /* CRYPTO_set_mem_ex_functions() needs a function rather than a macro */
+    /* unfortunately, OpenSSL provides no file:line information here */
+    str_free_debug(ptr, "OpenSSL", 0);
+}
+#endif
+
 /* end of str.c */
