@@ -1,4 +1,4 @@
-"""An error occured."""
+"""An error occurred."""
 
 if __name__ == '__main__':
     print("Wrong_connection!")
